//class declaration

//NODE CLASS
function Node (value) {
    this.value = Number(value);
    this.parent = "";
    this.children = [];
}

Node.prototype.addChild = function(newChild){

    if (this.children.length < 9){
        this.children.push(newChild);
    } else {
        var size = this.children.length;
        var lastChild = this.children[size-1];
        var preLastChild = this.children[size-2];

        lastChild.setParent(preLastChild);

        //try to get the children if there are not that much
        if(preLastChild.children.length + lastChild.children.length <= 9){
            preLastChild.adoptChildren(lastChild);
        }

        this.children.push(newChild);
        this.children.move(this.children.indexOf(newChild), 0);
    }
};

Node.prototype.realignChildren = function(){

    var mult = 10;
    if (this.value % 1 > 0){
        mult = Math.pow(10, this.value.countDecimals() +1);
    }

    for (var i = 0; i < this.children.length; i++) {
        this.children[i].value =
            ((this.value * mult) + (i+1)) / mult;

        if (this.children[i].children.length > 0 ){
            this.children[i].realignChildren();
        }
    }
};

Node.prototype.isPartentTakesChildren = function(){

    if (this.parent instanceof Node && this.parent.children.length < 9)
    {
        return true;
    }

    else {
        return false;
    }
};

Node.prototype.getChildPosition = function(child){
    return this.children.indexOf(child);
};

Node.prototype.getChildByValue = function(value){
    for (var i = 0; i < this.children.length; i++) {
        if( value == this.children[i].value){
            return this.children[i];
        }
    }

    for (var i = 0; i < this.children.length; i++){
        var potentialNode = this.children[i].getChildByValue(value);
        if (potentialNode instanceof Node){
            return potentialNode;
        }
    }
};

Node.prototype.toList = function(){
    var list = [];
    list.push(this.value);

    for (var i = 0; i < this.children.length; i++) {
        if (this.children[i].children.length > 0 ){
            list = list.concat(this.children[i].toList());
        } else {
            list.push(this.children[i].value);
        }
    }

    return list;
};

Node.prototype.removeChild = function(childNode){
    var childIndex = this.children.indexOf(childNode);
    if (childIndex >= 0){
        this.children.splice(childIndex, 1);
    }
};

Node.prototype.calcParentNumbers = function(){
    // console.log("calculate Parent Numbers for " + this.value);

    var decimals = this.value.countDecimals();

    if(decimals <= 1 && this.value % 1 === 0){
        return true;
    } else {
        var parentList = [];


        for (var i = 0; i < decimals; i++) {
            parentList.push(this.value.toFixedDown(i));
        }

        return parentList;
    }




    // var parentList = [];
    // if (isInt(this.value)){
    //     return "true";
    // } else {
    //     var decimals = this.value.countDecimals();
    //     parentList = [];
    //     for (var i = 0; i <= decimals; i++) {
    //         parentList.push(this.value.toFixedDown(i));
    //     }
    // }
    // return parentList;
};

Node.prototype.clearChildren = function(){
    this.children = [];
};

Node.prototype.setParent = function(newParent){
    if (this.parent instanceof Node ||
        this.parent instanceof NodeContainer){
            this.parent.removeChild(this);
    }

    this.parent = newParent;
    newParent.addChild(this);
};

Node.prototype.replacePosInParent = function(newEntry){
    newEntry.parent = this.parent;

    var pos = this.parent.getChildPosition(this);

    if (this.parent instanceof Node){
        this.parent.children[pos] = newEntry;
    } else if (this.parent instanceof NodeContainer){
        this.parent.contents[pos] = newEntry;
    }
};

Node.prototype.switchValues = function(node){
    var tmp = this.value;
    this.value = node.value;
    node.value = tmp;
};

Node.prototype.setFirstChildAsNewParent = function(){
    if (this.children.length > 0){
        var newParent = this.children[0];
        newParent.parent = this.parent;
        newParent.value = this.value;
        var thisPosInParent = this.parent.getChildPosition(this);

        if (this.parent instanceof Node){
            this.parent.children[thisPosInParent] = newParent;
        } else if (this.parent instanceof NodeContainer){
            this.parent.contents[thisPosInParent] = newParent;
        }

        for (var i = 1; i < this.children.length; i++) {
            newParent.addChild(this.children[i]);
            this.children[i].parent = newParent;
        }
        newParent.realignChildren();
        this.clearChildren();
    }
};

Node.prototype.adoptChildren = function(node){
    for (var i = 0; i < node.children.length; i++) {
        node.children[i].parent = this;
        this.addChild(node.children[i]);
    }
    node.clearChildren();
};

Node.prototype.changeParentalWithChild = function(child){

    this.switchValues(child);
    this.replacePosInParent(child);


    var childrenOfChild = child.children.slice();
    this.children[this.getChildPosition(child)] = this;
    child.clearChildren();
    child.adoptChildren(this);
    this.children = childrenOfChild;

};


//NOTE CONTAINER CLASS
function NodeContainer (){
    this.contents = [];
}

NodeContainer.prototype.nodeIsDirectChild = function(node){
    if (node.parent === this){
        return true;
    } else {
        return false;
    }
};

NodeContainer.prototype.calculateParental = function(node){

    // console.log("calculateParental for " + node);

    var parentList = node.calcParentNumbers();

    if (parentList === true){
        return this;
    } else {
        var actualParent;
        for (var i = 0; i < parentList.length; i++) {
            actualParent = this.getItemByValue(parentList[i]);

            if (actualParent instanceof Node){
                return actualParent;
            }
        }
    }

    // return this;
};

NodeContainer.prototype.readNewNode = function(newNode){

    var parent = this.calculateParental(newNode);

    if (parent instanceof Node){
        newNode.setParent(parent);
    } else if (parent instanceof NodeContainer){
        newNode.setParent(this);
    }
};

NodeContainer.prototype.addChild = function(child){
    this.contents.push(child);
};

NodeContainer.prototype.removeChild = function(childNode){
    var childIndex = this.contents.indexOf(childNode);

    if (childIndex >= 0){
        this.contents.splice(childIndex, 1);
    }
};

NodeContainer.prototype.getChildPosition = function(child){
    return this.contents.indexOf(child);
};

NodeContainer.prototype.getItemByValue = function(value){
    // console.log("get Item by value: " + value);

    for (var i = 0; i < this.contents.length; i++) {
        if (parseFloat(value) === this.contents[i].value) {
            return this.contents[i];
        }
    }

    var child = this.getItemByValue(Math.floor(value));
    return child.getChildByValue(value);
};

NodeContainer.prototype.toList = function(){
    var list = [];
    for (var i = 0; i < this.contents.length; i++) {
        var tList = this.contents[i].toList();
        list = list.concat(tList);
        // console.log("tList : " + i + " , " + list );
    }
    return list;
};

NodeContainer.prototype.switchNodePosition = function(nodeA, nodeB){
    var tmpIndexA = this.contents.indexOf(nodeA);
    var tmpIndexB = this.contents.indexOf(nodeB);
    this.contents[tmpIndexA] = nodeB;
    this.contents[tmpIndexB] = nodeA;
};

NodeContainer.prototype.setChildToTop = function(child){
    if (!(this.nodeIsDirectChild(child))){
        child.parent.removeChild(child);
        child.setParent(this);
    }

    var childPos = this.contents.indexOf(child);
    this.contents.move(childPos, 0);
};

//variable declaration
var fixHelper = function(e, ui) {
    ui.children().each(function() {
        $(this).width($(this).width());
    });
    return ui;
};

//execute

$( document ).ready(function() {

    var positions = getTableOrderAsList();
    var rootNode = new NodeContainer();

    for (var i = 0; i < positions.length; i++) {
        rootNode.readNewNode(new Node(positions[i]
            .getElementsByTagName("td")[1].innerHTML));
    }


    $("table tbody").sortable({
        items: ".list-entrys",
        placeholder: 'marker',
        revert: 100,
        helper: fixHelper,

        update: function( event, ui ) {
            positions = getTableOrderAsList();

            var movedItem = ui.item[0];

            var movedItemIndex = positions.indexOf(movedItem);
            var parentItemIndex = positions.indexOf(movedItem) -1;
            var childItemIndex = positions.indexOf(movedItem) +1;

            var parentNode;
            var childNode;
            var itemNode = rootNode.
                getItemByValue(positions[movedItemIndex]
                .getElementsByTagName("td")[1].innerHTML);

            console.log("par" + parentItemIndex);

            if (parentItemIndex >= 0){
                parentNode = rootNode.getItemByValue(
                    positions[parentItemIndex]
                    .getElementsByTagName("td")[1].innerHTML
                );
            }

            if (childItemIndex < positions.length){
                childNode = rootNode.getItemByValue(
                    positions[childItemIndex]
                    .getElementsByTagName("td")[1].innerHTML
                );
            }

            if (parentNode instanceof Node &&
                childNode instanceof Node ||
                parentNode instanceof Node &&
                !(childNode instanceof Node))
            {
                // console.log("both are defined");

                if (itemNode === parentNode.parent){
                    itemNode.changeParentalWithChild(parentNode);
                    parentNode.realignChildren();
                }

                else if (parentNode.parent === itemNode.parent &&
                    itemNode.parent instanceof Node)
                {
                    parentNode.parent.realignChildren();
                }

                else if (parentNode.isPartentTakesChildren()){
                    itemNode.setFirstChildAsNewParent();
                    itemNode.setParent(parentNode.parent);
                    parentNode.parent.realignChildren();
                }

                else {
                    itemNode.setFirstChildAsNewParent();
                    itemNode.setParent(parentNode);
                    parentNode.realignChildren();
                }
            }

            else if (childNode instanceof Node &&
                !(parentNode instanceof Node))
            {
                console.log("very fist element");
                itemNode.setFirstChildAsNewParent();
                itemNode.setParent(rootNode);

                // move the item on the first place
                rootNode.setChildToTop(itemNode);

                //add the child to the parent
                childNode.setParent(itemNode);

                //adopt the children of the child
                itemNode.adoptChildren(childNode);

                //get the value of the next item
                itemNode.value = Math.floor(childNode.value);

                itemNode.realignChildren();
            }

            // write back
            var newList = rootNode.toList();
            for (var i = 0; i < positions.length; i++) {
                var entry = positions[i].getElementsByTagName("td")[1];
                var oldentry = entry.innerHTML.slice();
                entry.innerHTML = newList[i].roundupNumber();

                if (entry.innerHTML !== oldentry){
                    console.log("writeBackInDB: " + entry.innerHTML);

                    //get the URL of the issue
                    var issuenum = positions[i].getElementsByTagName("td")[3]
                        .getElementsByTagName("a")[0].getAttribute("href");

                    postToRoundup(issuenum, entry.innerHTML);

                    console.log("BP");

                }
            }
            // console.log("bp");
        },
    });
});

//functions

function postToRoundup(issuenum, sequenzeNumber){
    // get old issue data


    $.ajax({
        type: 'POST',
        url: issuenum,
        async: true,
        success: function (data, textStatus, jqXHR){
            var code = $(data);
            var input_dict = {};
            var fd = new FormData();
            var date = new Date();
            var file = new File([""], "filename");
            var typeFields = code.find("input");
            var selectFields = code.find("select");

            for (var i = 0; i < typeFields.length; i++) {
                input_dict[typeFields[i].name] = typeFields[i].value;
            }

            for (var i = 0; i < selectFields.length; i++) {
                input_dict[selectFields[i].name] = selectFields[i].value;
            }

            fd.append('title', input_dict.title);
            fd.append('priority', input_dict.priority);
            fd.append('status', input_dict.status);
            fd.append('superseder', input_dict.superseder);
            fd.append('nosy', input_dict.nosy);
            fd.append('assignedto', input_dict.assignedto);
            fd.append('keyword', "");
            fd.append('sequenze', sequenzeNumber);
            fd.append('@note', 'change sequence with R-DND');
            fd.append('@file', file, "");
            fd.append('@template', 'item');
            fd.append('@required', 'title,priority');
            fd.append('@lastactivity', date
                .toISOString()
                .replace("T", ".")
                .slice(0,19));
            fd.append('@action', 'edit');
            fd.append('submit_button', 'Submit Changes');

            $.ajax({
                type: "POST",
                url: issuenum,
                headers: {
                    'Cache-Control': 'max-age=0',
                    'Accept': 'text/html,' +
                        'application/xhtml+xml,' +
                        'application/xml;q=0.9,' +
                        'image/webp,' +
                        '*/*;q=0.8'
                },
                async: false,
                data: fd,
                contentType: false,
                processData: false,
            });
        }
    });
}


function getTableOrderAsList(){
    var positions = [];

    var rows = document.getElementsByClassName('list')[0].
        getElementsByTagName("tbody")[0].rows;

    for (var i = 0; i < rows.length; i++){

        if ($(rows[i]).hasClass('list-entrys') ){
            positions.push(rows[i]);
        }
    }

    return positions;
}


if (typeof String.prototype.startsWith != 'function') {
  // see below for better implementation!
  String.prototype.startsWith = function (str){
    return this.indexOf(str) === 0;
  };
}


Number.prototype.roundupNumber = function(){
    if (this.valueOf() % 1 == 0){
        return ""+this+".0";
    } else {
        return ""+this;
    }
};


Number.prototype.countDecimals = function() {
    if ( !isNaN(this.valueOf())){

        if(Math.floor(this.valueOf()) === this.valueOf()){
            return 0;
        } else {
            return this.toString().split(".")[1].length || 0;
        }

    }
};


Number.prototype.toFixedDown = function(digits) {
    var re = new RegExp("(\\d+\\.\\d{" + digits + "})(\\d)"),
        m = this.toString().match(re);
    return m ? parseFloat(m[1]) : this.valueOf();
};


Number.prototype.cutLastNumber = function(){
    if (this instanceof Number){
        return this.toFixedDown(this.countDecimals() -1);
    }
};


Array.prototype.move = function (old_index, new_index) {
    if (new_index >= this.length) {
        var k = new_index - this.length;
        while ((k--) + 1) {
            this.push(undefined);
        }
    }
    this.splice(new_index, 0, this.splice(old_index, 1)[0]);
    return this; // for testing purposes
};
